package com.cbnanashi.entitybackup;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;
import org.apache.commons.httpclient.DefaultHttpMethodRetryHandler;
import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpException;
import org.apache.commons.httpclient.HttpStatus;
import org.apache.commons.httpclient.NameValuePair;
import org.apache.commons.httpclient.methods.GetMethod;
import org.apache.commons.httpclient.methods.PostMethod;
import org.apache.commons.httpclient.methods.multipart.FilePart;
import org.apache.commons.httpclient.methods.multipart.MultipartRequestEntity;
import org.apache.commons.httpclient.methods.multipart.Part;
import org.apache.commons.httpclient.params.HttpMethodParams;

public class EntityRestore {
	private static final String BACKUP_FILE_SUFFIX = ".txt";
	private static final String KEY_INDICATOR = "__key__";
	private static final String KIND_DELETE_PARAM_NAME = "deleteKind";

	private static final int MAX_NORMAL_POST_SIZE = 1024 * 128; // multipartł͂ȂʏPOSTőMf[^̍őTCYA128KB
	private static final int SERVER_ERROR_RETRY_COUNT = 5;
	private static final int SERVER_ERROR_WAIT_SECONDS = 10;
	private static final String TEMP_FILE_PREFIX = "EntityRestoreTemp";
	private static final int INITIAL_CHUNK_SIZE = 1024 * 32; // ɑMTCY̏lA32KBiGeBeB̂قe傫悤j

	private String servletUrl;
	private List<String> kinds;
	private int minimumChunkSize = INITIAL_CHUNK_SIZE;
	private List<String> entityStrings = new LinkedList<String>();
	private int retryCount = 0;
	private boolean deleteBeforeRestore;

	// RXgN^
	public EntityRestore(boolean delete, String u, List<String> l) {
		deleteBeforeRestore = delete;
		servletUrl = u;
		kinds = l;
	}

	// 1:-dō폜RestoreA+1:URLAȍ~:KindiIvVj
	public static void main(String[] args) {
		if (args.length == 0) {
			System.out.println("Usage:\tjava -jar entityrestore.jar [-d] url [kind...]");
			System.exit(99);
		}

		boolean delete = false;

		List<String> argList = new LinkedList<String>(Arrays.asList(args));
		String arg0 = argList.remove(0);

		if (arg0.equals("-d")) {
			delete = true;
			arg0 = argList.remove(0);
		}
		String url = arg0;

		EntityRestore restorer = new EntityRestore(delete, url, argList);

		try {
			restorer.restore();
		} catch (Exception e) {
			e.printStackTrace();
			System.exit(99);
		}
		System.exit(0);
	}

	// restore
	private void restore() {
		log("URL̃AvP[VpăXgA܂BURL:" + servletUrl);

		// KINDw肳ĂȂΓǂݍ
		if (kinds == null || kinds.isEmpty()) {
			loadKinds();
		}

		StringBuilder sb = new StringBuilder();
		for (String kind : kinds) {
			sb.append(kind).append('\n');
		}
		log("KindΏۂƂ܂B\n" + sb.toString());

		// KindƂɏ
		for (String kind : kinds) {
			restoreKind(kind);
		}

		log("܂B");
	}

	// kindrestore
	private void restoreKind(String kind) {
		// 폜[hȂ폜
		if (deleteBeforeRestore) {
			deleteKind(kind);
		}

		// ̓t@CJ
		BufferedReader in;
		try {
			in = new BufferedReader(new FileReader(kind + BACKUP_FILE_SUFFIX));
		} catch (IOException e) {
			throw new IllegalStateException("t@CJ܂B", e);
		}

		log("KindBKind:" + kind);
		int count = 0;
		try {
			String line;
			List<String> lines = new LinkedList<String>();
			for (;;) {
				line = in.readLine();
				if (!lines.isEmpty() && (line == null || line.startsWith(KEY_INDICATOR))) {
					// ꌏ̃f[^𑗐M
					addEntity(lines);
					count++;
					lines.clear();
				}
				if (line == null) {
					break;
				}
				lines.add(line);
			}
			checkChunk(true);
		} catch (IOException e) {
			throw new IllegalStateException("t@C̓G[łB", e);
		} finally {
			try {
				in.close();
			} catch (IOException e) {
				// ignore
			}
		}

		log("Kind:" + kind + "" + count + "XgA܂B");
	}

	// Entityobt@ɒǉ
	private void addEntity(List<String> lines) throws IOException {
		StringBuffer sb = new StringBuffer();
		for (String line : lines) {
			sb.append(line).append('\n');
		}
		entityStrings.add(sb.toString());
		checkChunk(false);
	}

	// chunk܂擪YTCYM
	private void checkChunk(boolean flush) throws IOException {
		if (entityStrings.isEmpty()) {
			return;
		}

		List<String> chunk = new ArrayList<String>();
		int size = 0;
		for (String entityString : entityStrings) {
			chunk.add(entityString);

			size += entityString.getBytes("UTF-8").length;
			if (size > minimumChunkSize) {
				break;
			}
		}
		if (!flush && size <= minimumChunkSize) {
			return; // Ȃ
		}

		// ܂
		log("T[oɑM܂Bcount:" + chunk.size() + ",size:" + size);
		boolean success = restoreEntities(chunk);
		if (!success) {
			// T[oG[̓gC
			retryCount++;
			if (retryCount == SERVER_ERROR_RETRY_COUNT) {
				throw new IllegalStateException("Server Error" + SERVER_ERROR_RETRY_COUNT + "AĔ܂B");
			}

			log("T[oG[܂BgC܂Bcount:" + retryCount);
			try {
				Thread.sleep(SERVER_ERROR_WAIT_SECONDS * 1000);
			} catch (InterruptedException e) {
				// ignore
			}
			minimumChunkSize /= 2; // TCYăgC
			checkChunk(flush);
			return;
		}

		// 
		retryCount = 0;
		for (int i = 0; i < chunk.size(); i++) {
			entityStrings.remove(0);
		}
		checkChunk(flush);
	}

	private boolean restoreEntities(List<String> restoreStrings) {
		StringBuffer sb = new StringBuffer();
		for (String s : restoreStrings) {
			sb.append(s);
		}
		String content = sb.toString();

		PostMethod post = null;
		File temp = null;
		try {
			HttpClient client = new org.apache.commons.httpclient.HttpClient();

			post = new PostMethod(servletUrl);

			byte[] bytes = content.getBytes("UTF-8");
			if (bytes.length <= MAX_NORMAL_POST_SIZE) {
				// ʏPOSTőM
				post.addRequestHeader("Content-Type", "application/x-www-form-urlencoded; charset=UTF-8");
				NameValuePair[] data = { new NameValuePair("data", content) };
				post.setRequestBody(data);
			} else {
				// file upload(multipart)őM
				// e|t@Cɏo͂
				try {
					temp = File.createTempFile(TEMP_FILE_PREFIX, null);
					OutputStream out = new FileOutputStream(temp);
					out.write(bytes); // http client̓oCiőM̂UTF-8ŏo͂Ă
					out.close();
				} catch (IOException e) {
					throw new IllegalStateException("e|t@Cւ̏o̓G[łB", e);
				}

				Part[] parts = { new FilePart("data", temp) };
				post.setRequestEntity(new MultipartRequestEntity(parts, post.getParams()));
			}
			post.getParams().setParameter(HttpMethodParams.RETRY_HANDLER, new DefaultHttpMethodRetryHandler(3, false));

			int status = client.executeMethod(post);
			if (status == HttpStatus.SC_OK) {
				return true; // OK
			} else if (status != HttpStatus.SC_INTERNAL_SERVER_ERROR) {
				throw new IllegalStateException("XgÃG[łBstatus:" + status);
			}

			// T[oG[̓gC
			return false;
		} catch (HttpException e) {
			throw new IllegalStateException("vgRG[łB", e);
		} catch (IOException e) {
			throw new IllegalStateException("]G[łB", e);
		} finally {
			if (post != null) {
				post.releaseConnection();
			}
			if (temp != null) {
				temp.delete();
			}
		}
	}

	private void log(String s) {
		System.out.println(s);
	}

	// JgfBNgt@CꗗKind̎ނ擾
	private void loadKinds() {
		File[] files = (new File(".")).listFiles(new FilenameFilter() {
			public boolean accept(File dir, String name) {
				return name.endsWith(BACKUP_FILE_SUFFIX);
			}
		});

		kinds = new LinkedList<String>();
		for (File f : files) {
			String name = f.getName();
			kinds.add(name.substring(0, name.lastIndexOf('.')));
		}
	}

	private String encode(String s) {
		try {
			return URLEncoder.encode(s, "UTF-8");
		} catch (UnsupportedEncodingException e) {
			// NȂ͂
			return s;
		}
	}

	private void deleteKind(String kind) {
		String url = servletUrl + "?" + KIND_DELETE_PARAM_NAME + "=" + encode(kind);

		HttpClient client = new HttpClient();
		for (;;) {
			GetMethod get = null;
			try {
				log("Kind폜BKind:" + kind);

				get = new GetMethod(url);

				int status = client.executeMethod(get);
				if (status == HttpStatus.SC_NOT_FOUND) {
					break; // I
				} else if (status == HttpStatus.SC_SERVICE_UNAVAILABLE) {
					log("T[oG[i503j߁A΂炭wait܂B");
					try {
						Thread.sleep(SERVER_ERROR_WAIT_SECONDS * 1000);
					} catch (InterruptedException e) {
						// ignore
					}
				} else if (status != HttpStatus.SC_OK) {
					throw new IllegalStateException("Kind폜̃G[łBstatus:" + status);
				}
			} catch (HttpException e) {
				throw new IllegalStateException("vgRG[łB", e);
			} catch (IOException e) {
				throw new IllegalStateException("ʐMG[łB", e);
			} finally {
				if (get != null) {
					get.releaseConnection();
				}
			}
		}
	}

}
